import os
import importlib
import collections
from . import sd

__version__ = '0.6.6'

# Name and description.
# Note that the name should match to the module name, gen_<name>.
SUPPORTED_FORMATS = collections.OrderedDict()
SUPPORTED_FORMATS['plantuml'] = 'PlantUML script (default)'
SUPPORTED_FORMATS['plantuml_png'] = 'PlantUML script and PNG image'
SUPPORTED_FORMATS['plantuml_svg'] = 'PlantUML script and SVG image'
SUPPORTED_FORMATS['plantuml_txt'] = 'PlantUML script and ASCII art text'
DEFAULT_FORAMT = 'plantuml'

_collected_seq_diagrams = []


class seq_diagram:
    """
    Decorator to mark the sequence  function.

    'diagram_name' is the file name for the generated . The name of
    decorated function will be used if not specified.

    ex:
    @napkin.seq_diagram()
    def sd_simple(c):
       ...

    @napkin.seq_diagram('sd_another')
    def foo(c):
       ...
    """
    def __init__(self, name=None):
        self.name = name

    def __call__(self, wrapped_func):
        if not self.name:
            self.name = wrapped_func.__name__
        self.sd_func = wrapped_func
        _collected_seq_diagrams.append(self)


def generate(output_format=DEFAULT_FORAMT, output_dir='.', options=None):
    """
    Generate sequence diagrams from all the decorated functions.
    """
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    module_name = '.gen_' + output_format
    gen_module = importlib.import_module(module_name, 'napkin')

    for d in _collected_seq_diagrams:
        context = sd.parse(d.sd_func)
        generated_files = gen_module.generate(d.name, output_dir, context,
                                              options if options else {})
        print('File generated : {}'.format(', '.join(generated_files)))



"""
Command line interface to generate sequence diagrams.
"""
import os
import argparse
import re
from . import generate, SUPPORTED_FORMATS, DEFAULT_FORAMT, __version__

_DESCRIPTION = 'Generate UML sequence diagram from Python code'
_EPILOG = """
Supported output formats:
""" + '\n'.join('  {:16} : {}'.format(k, v)
                for k, v in SUPPORTED_FORMATS.items())


def _parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=_DESCRIPTION,
        epilog=_EPILOG)

    parser.add_argument(
        '--output-format', '-f',
        default=DEFAULT_FORAMT, choices=SUPPORTED_FORMATS.keys()),
    parser.add_argument(
        '--output-dir', '-o', default='.')
    parser.add_argument(
        'srcs', nargs='+',
        help='Python file or directory containing diagram functions')
    parser.add_argument(
        '--version', action='version', version=__version__)

    #
    # Format specific arguments. It is not desirable to have here and it will
    # be refactored once we have many of them.
    #
    parser.add_argument(
        '--server-url', default=argparse.SUPPRESS,
        help='(only for plantuml_png/svg format) Default is the public server')

    return parser.parse_args()


def _import_script(fname):
    with open(fname, 'rt') as f:
        file_contents = f.read()
        if '@napkin.seq_diagram' in file_contents:
            print('Load file : {}'.format(fname))
            exec(compile(file_contents, fname, 'exec'), globals(), locals())


def _collect_py_files(srcs):
    collected = []
    for src in srcs:
        if os.path.isdir(src):
            for root, dirs, files in os.walk(src):
                collected += [os.path.join(root, f)
                              for f in files if re.match(r'\w*\.py$', f)]
        else:
            collected.append(src)
    return collected


def main():
    args = _parse_args()
    for fname in _collect_py_files(args.srcs):
        _import_script(fname)
    generate(args.output_format, args.output_dir, options=vars(args))



"""
Sequence diagram elements and API
"""

import collections
from . import sd_action


class ContextError(Exception):
    pass


class TopLevelCallerError(ContextError):
    pass


class ReturnError(ContextError):
    pass


class CallError(ContextError):
    pass


class FragError(ContextError):
    pass


class CreateError(ContextError):
    pass


class DestroyError(ContextError):
    pass


class Params:
    def __init__(self, args=None, kargs=None):
        self.args = args if args else None
        self.kargs = kargs if kargs else None

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __str__(self):
        s = ''
        if self.args:
            s += ', '.join(['%s' % str(a) for a in self.args])
            if self.kargs:
                s += ', '
        if self.kargs is not None:
            s += ', '.join(['%s=%s' % (k, v) for k, v in self.kargs.items()])
        return s


class MethodCall:
    def __init__(self, obj, method, args, kargs):
        self.obj = obj
        self.method = method
        self.params = Params(args, kargs)

        #
        # ret_params can be Params object to specify explicit return value.
        # None means implicit return
        #
        self.ret_params = None
        self.args = args
        self.kargs = kargs
        self.notes = [None, None]

    def __enter__(self):
        self.obj.enter_call(self)

    def __exit__(self, *exc_args):
        if exc_args[0]:
            return
        self.obj.exit_call()

    def ret(self, *args, **kargs):
        """
        Specify return params.
        ex)
        foo.func().return('value')
        """
        self.specify_return_params(Params(args, kargs))

    def note(self, callee=None, caller=None):
        """
        Specify note for caller/callee side.
        """
        assert callee or caller, 'At least one argument necessary'
        self.notes[:] = (callee, caller)
        return self

    def specify_return_params(self, params):
        if self.ret_params:
            raise ReturnError('double return')
        self.ret_params = params


class Method:
    def __init__(self, obj, name, flags=''):
        self.obj = obj
        self.name = name
        self.flags = flags

    def __call__(self, *args, **kargs):
        call = MethodCall(self.obj, self, args, kargs)
        self.obj.invoke_call(call)
        return call

    def __str__(self):
        return self.name


class Object(object):
    def __init__(self, sd, name, cls=None, stereotype=None, instance_id=None):
        self.sd = sd
        self.name = name
        self.cls = cls
        self.methods = {}
        self.stereotype = stereotype

        # Becomes false if it is destroyed in the middle of sequence diagram
        self.valid = True
        # Becomes true if it is created in the middle of sequence diagram
        self.created = False
        self.instance_id = instance_id

    def __getattr__(self, name):
        return self.create_method(name)

    def create_method(self, name):
        method = self.methods.setdefault(name, Method(self, name))
        return method

    def __enter__(self):
        self.sd.enter_top_object(self)

    def __exit__(self, *exc_args):
        if exc_args[0]:
            return
        self.sd.leave_top_object()

    def __repr__(self):
        return self.name

    def invoke_call(self, call):
        self.sd.invoke_call(call)

    #
    # Used by MethodCall
    #
    def enter_call(self, call):
        self.sd.enter_call(call)

    def exit_call(self):
        self.sd.exit_call()

    def note(self, text):
        self.sd.note_over(self, text)


class Frag:
    def __init__(self, sd, op_name, condition=None):
        self.sd = sd
        self.op_name = op_name
        self.condition = condition

    def __enter__(self):
        self.sd.enter_frag(self)

    def __exit__(self, *exc_args):
        if exc_args[0]:
            return
        self.sd.exit_frag()


class Context(object):
    """
    Context to give API to the user diagram function and it also captures the
    diagram representation.
    """
    def __init__(self):
        # Objects used in the diagram.
        self._objects = collections.OrderedDict()

        # Capture the sequence of the actions
        self._sequence = []

        self._call_stack = []
        self._current_call = None

        self._frag_stack = []
        self._current_frag = None

        # Call created by 'obj.func' can be used both with/without 'with'.
        # self._pending_call remembers it to add proper 'return' if necessary.
        # For example,
        # ex)
        #     ...
        #     foo.func                # (1)
        #     foo.func2               # causes to add 'implicit return' for (1)
        #
        # ex)
        #     with foo.func():        # (1)
        #         bar.func            # (2)
        #
        #     When (1) exits, pending_call associated with (2) to add 'implicit
        #     return' for (2)
        #
        self._pending_call = None
        self._num_objects = 0

    def object(self, name, cls=None, stereotype=None):
        """Create an object
        """
        obj = self._objects.setdefault(name,
                                       Object(self, name, cls=cls,
                                              stereotype=stereotype,
                                              instance_id=self._num_objects))
        self._num_objects += 1
        return obj

    def enter_top_object(self, obj):
        if self._call_stack:
            raise TopLevelCallerError('Top level caller cannot be set again')
        assert self._current_call is None

        self._call_stack.append(self._current_call)
        # Add a guard call as the first item, which makes the overall logic
        # simpler.
        guard_call = MethodCall(obj, Method(obj, '<<somewhere>>'), None, None)
        self._current_call = guard_call

    def leave_top_object(self):
        assert len(self._call_stack) == 1
        self.return_any_pending_call()

        self._current_call = self._call_stack.pop()
        assert self._current_call is None

    def invoke_call(self, call):
        if self._current_call.ret_params:
            raise CallError('Current function already returned')

        if self._current_frag and self._current_frag.op_name == 'alt':
            raise CallError('Cannot be invoked inside alt level')

        self.return_any_pending_call()

        if not call.obj.valid:
            raise CallError('Object was destroyed')

        self._pending_call = call

    def return_any_pending_call(self):
        pending_call = self._pending_call
        if pending_call:
            caller = self._current_call.obj
            call_action = sd_action.Call(caller,
                                         pending_call.obj,
                                         pending_call.method.name,
                                         pending_call.params,
                                         pending_call.method.flags,
                                         pending_call.notes)
            self._sequence.append(call_action)

            ret_params = self._pending_call.ret_params
            ret_action = (sd_action.Return(ret_params)
                          if ret_params else
                          sd_action.ImplicitReturn())
            self._add_return_action(self._pending_call, ret_action)
            self._pending_call = None

    def _add_return_action(self, call, ret_action):
        if 'd' in call.method.flags:

            # Make it invalid to use the object after the destructor returns.
            call.obj.valid = False

        self._sequence.append(ret_action)

    def enter_call(self, call):
        caller = self._current_call.obj
        action = sd_action.Call(caller, call.obj,
                                call.method.name, call.params,
                                call.method.flags, call.notes)

        self._sequence.append(action)

        self._call_stack.append(self._current_call)
        self._current_call = call
        self._pending_call = None

    def exit_call(self):
        #
        # ex)
        # ...
        # with foo.func():     (2)
        #     bar.func()       (1)
        #
        # for returning (1)
        self.return_any_pending_call()
        # for returning (2)
        ret_params = self._current_call.ret_params
        ret_action = (sd_action.Return(ret_params)
                      if ret_params else
                      sd_action.ImplicitReturn())
        self._add_return_action(self._current_call, ret_action)

        self._current_call = self._call_stack.pop()

    def ret(self, *args, **kargs):
        """
        Return from the current call
        """
        if len(self._call_stack) == 1:
            """
            Invalid operation in the top level caller
            ex)
                    with foo:
                        bar.func()
                        c.ret()             <--- Invalid
            """
            raise CallError('Return from top level caller')

        """
        ex)
                with foo.func():
                    bar.func()
                    c.ret()             <---
        """
        self._current_call.specify_return_params(Params(args, kargs))

    def opt(self, condition=None):
        return Frag(self, 'opt', condition)

    def loop(self, condition=None):
        return Frag(self, 'loop', condition)

    def alt(self):
        return Frag(self, 'alt')

    def choice(self, condition=None):
        return Frag(self, 'choice', condition)

    def group(self, label=None):
        return Frag(self, 'group', label)

    def enter_frag(self, frag):
        if self._current_call and self._current_call.ret_params:
            raise FragError('Current function already returned')

        self.return_any_pending_call()

        if self._current_frag and self._current_frag.op_name == 'alt':
            if frag.op_name != 'choice':
                raise FragError('Other frag except choice not '
                                'allowed inside alt')

        self._frag_stack.append(self._current_frag)
        self._current_frag = frag
        self._sequence.append(
            sd_action.FragBegin(frag.op_name, frag.condition))

    def exit_frag(self):
        self.return_any_pending_call()

        # Check if there is no action inside frag
        last_sequence = self._sequence[-1]
        if isinstance(last_sequence, sd_action.FragBegin):
            raise FragError('Empty fragment')

        self._sequence.append(sd_action.FragEnd(self._current_frag.op_name))
        self._current_frag = self._frag_stack.pop()

    def create(self, obj_or_call):
        """Create object and invoke the constructor.

        If obj_or_call is Object type, default constructor called, <<create>>
        is called.
        """

        if self._current_call is None:
            raise CreateError('No caller specified')

        if isinstance(obj_or_call, Object):
            obj = obj_or_call
            method = obj.create_method('<<create>>')
            call = method()
        elif isinstance(obj_or_call, MethodCall):
            call = obj_or_call
            obj = call.obj
            method = call.method
        else:
            raise CreateError('Invalid argument')

        if len(obj.methods) > 1:
            raise CreateError('Object is already being used')

        if obj.created:
            raise CreateError('Object was already created')
        else:
            obj.created = True

        method.flags = 'c'
        return call

    def destroy(self, obj_or_call):
        """Invoke the destructor then destroy the object.
        """
        if self._current_call is None:
            raise DestroyError('No caller specified')

        if isinstance(obj_or_call, Object):
            obj = obj_or_call
            if not obj.valid:
                raise DestroyError('Object already destroyed')
            method = obj.create_method('<<destroy>>')
            call = method()
        elif isinstance(obj_or_call, MethodCall):
            call = obj_or_call
            obj = call.obj
            method = call.method
        else:
            raise CreateError('Invalid argument')

        # the object will become invalid when it returns from the destructor
        method.flags = 'd'
        return call

    def note(self, text):
        """Add note over current object.
        """
        if self._current_call is None:
            raise DestroyError('No object specified')
        self._current_call.obj.note(text)

    def note_over(self, obj, text):
        self.return_any_pending_call()
        self._sequence.append(sd_action.Note(text, obj=obj))

    def delay(self, text=None):
        self.return_any_pending_call()
        self._sequence.append(sd_action.Delay(text))


def parse(sd_func):
    """
    Parse sequence diagram.
    """
    sd_context = Context()
    sd_func(sd_context)
    return sd_context



import os
import argparse

from .gen_plantuml_img import generate_image

_DESCRIPTION = ('Simple tool to convert PlantUML text file into image file '
                'using server.')


def _parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=_DESCRIPTION)

    parser.add_argument('--server-url', help='Default is the public server')
    parser.add_argument('input_file', help='PlantUML text file ')
    parser.add_argument(
        'output_file',
        help=('Image file to generate. '
              'Note that the extension name decides the format.'))
    return parser.parse_args()


def main():
    args = _parse_args()
    _, image_type = os.path.splitext(args.output_file)
    generate_image(args.input_file, args.output_file, args.server_url)



import os
import napkin
from napkin.gen_plantuml_img import _encode_text_diagram

DEFAULT_SERVER_URL = 'http://www.plantuml.com/plantuml'


def test_encode_text_diagram():
    text = """@startuml
Bob -> Alice : hello
@enduml"""
    exp = b'SoWkIImgAStDuNBAJrBGjLDmpCbCJbMmKiX8pSd9vt98pKi1IW80'
    act = _encode_text_diagram(text)
    assert exp == act


def test_png_generation(tmpdir):
    fname = os.path.join(str(tmpdir), 'sd')

    @napkin.seq_diagram(fname)
    def f(c):
        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()

    napkin.generate(output_format='plantuml_png')
    assert os.stat(fname + '.puml')
    assert os.stat(fname + '.png')


def test_svg_generation(tmpdir):
    fname = os.path.join(str(tmpdir), 'sd')

    @napkin.seq_diagram(fname)
    def f(c):
        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()

    napkin.generate(output_format='plantuml_svg')
    assert os.stat(fname + '.puml')
    assert os.stat(fname + '.svg')



import pytest

from napkin import sd
from napkin import sd_action


class TestBase(object):
    def check(self, context, exp_actions):
        actions = context._sequence

        # This is for better debugging
        assert str(actions) == str(exp_actions)
        assert actions == exp_actions


class TestOpt(TestBase):
    def test_top_level(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with c.opt():
            with foo:
                bar.func()

        self.check(c, [
            sd_action.FragBegin('opt'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
        ])

    def test_top_level_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with c.opt():
            with foo:
                bar.func()

        with c.opt():
            with foo:
                bar.func()

        self.check(c, [
            sd_action.FragBegin('opt'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
            sd_action.FragBegin('opt'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
        ])

    def test_inside_top_level(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with foo:
            with c.opt():
                bar.func()

        self.check(c, [
            sd_action.FragBegin('opt'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
        ])

    def test_inside_top_level_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with foo:
            with c.opt():
                bar.func()
            with c.opt():
                bar.func()

        self.check(c, [
            sd_action.FragBegin('opt'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
            sd_action.FragBegin('opt'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
        ])

    def test_inside_func_call(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')

        with foo:
            with bar.func():
                with c.opt():
                    baz.func()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.FragBegin('opt'),
            sd_action.Call(bar, baz, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
            sd_action.ImplicitReturn(),
        ])

    def test_inside_func_call_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')

        with foo:
            with bar.func():
                with c.opt():
                    baz.func()
                with c.opt():
                    baz.func()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.FragBegin('opt'),
            sd_action.Call(bar, baz, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
            sd_action.FragBegin('opt'),
            sd_action.Call(bar, baz, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
            sd_action.ImplicitReturn(),
        ])

    def test_with_condition(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with c.opt('condition'):
            with foo:
                bar.func()

        self.check(c, [
            sd_action.FragBegin('opt', 'condition'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('opt'),
        ])

    def test_fail_after_return_from_func(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with foo:
            with bar.func():
                c.ret()
                with pytest.raises(sd.FragError):
                    with c.opt():
                        pass

    def test_fail_if_empty_inside(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with foo:
            with bar.func():
                with pytest.raises(sd.FragError):
                    with c.opt():
                        pass

    def test_exception_inside(self):
        c = sd.Context()

        foo = c.object('foo')

        with pytest.raises(ZeroDivisionError):
            with foo:
                with c.opt():
                    3 / 0  # Cause exception


class TestLoop(TestBase):
    def test_top_level(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with c.loop():
            with foo:
                bar.func()

        self.check(c, [
            sd_action.FragBegin('loop'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('loop'),
        ])


class TestAlt(TestBase):
    def test_top_level_two_choices(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')

        with c.alt():
            with c.choice('ok'):
                with foo:
                    bar.func()
            with c.choice('else'):
                with foo:
                    baz.func()
        self.check(c, [
            sd_action.FragBegin('alt'),
            sd_action.FragBegin('choice', 'ok'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('choice'),
            sd_action.FragBegin('choice', 'else'),
            sd_action.Call(foo, baz, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('choice'),
            sd_action.FragEnd('alt'),
        ])

    def test_fail_if_call_in_alt(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with pytest.raises(sd.CallError):
            with c.alt():
                with foo:
                    bar.func()  # need to be inside choice

    def test_fail_if_other_frag_in_alt(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with pytest.raises(sd.FragError):
            with c.alt():
                with c.opt():  # wrong frag
                    with foo:
                        bar.func()

    def test_inside_top_level_two_sections(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')

        with c.alt():
            with c.choice('ok'):
                with foo:
                    bar.func()
            with c.choice('else'):
                with foo:
                    baz.func()

        self.check(c, [
            sd_action.FragBegin('alt'),
            sd_action.FragBegin('choice', 'ok'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('choice'),
            sd_action.FragBegin('choice', 'else'),
            sd_action.Call(foo, baz, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('choice'),
            sd_action.FragEnd('alt'),
        ])


class TestGroup(TestBase):
    def test_top_level(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')

        with c.group():
            with foo:
                bar.func()

        self.check(c, [
            sd_action.FragBegin('group'),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.FragEnd('group'),
        ])



"""
PlanUML format sequence diagram generation
"""

import re
import os
from . import sd_action
from . import util


def _output_participants(sd_context):
    """
    Generate a string containing participants in order of the occurrence.
    """
    output = []
    for o in sd_context._objects.values():
        stereotype = ' <<{}>>'.format(o.stereotype) if o.stereotype else ''
        output.append(
            'participant "{name:s}:{cls:s}" as {name:s}{stereotype}'.format(
                name=o.name, cls=o.cls, stereotype=stereotype)
            if o.cls else
            'participant {name:s}{stereotype}'.format(
                name=o.name, stereotype=stereotype))
    output.append('')
    return output


def _is_caller_left_side(caller, callee):
    # The first object is on the most left side. Note that for self-call,
    # caller is considered to be left side.
    return caller.instance_id <= callee.instance_id


def _generate_note(output, direction, obj, text):
    lines = text.splitlines()
    header = 'note ' + ('over {}'.format(obj) if obj else direction)
    if len(lines) > 1:
        output.append(header)
        output += lines
        output.append('end note')
    else:
        output.append('{} : {}'.format(header, text))


def _generate_script(sd_context):
    """
    Generate a string containing PlanUML script.
    """
    call_stack = []
    current_call = None

    output = []
    output.append('@startuml')
    output += _output_participants(sd_context)

    for p_action, action, n_action in util.neighbour(sd_context._sequence):
        if isinstance(action, sd_action.Call):
            if 'c' in action.flags:
                output.append('create %(callee)s' % action.__dict__)

            if re.match(r'<<\w+>>', action.method_name):
                # such as <<create>> or <<destroy>>
                output.append('%(caller)s -> %(callee)s : '
                              '%(method_name)s' % action.__dict__)
            else:
                output.append('%(caller)s -> %(callee)s : '
                              '%(method_name)s(%(params)s)' % action.__dict__)
            if action.notes != [None, None]:
                caller_side, callee_side = (
                    ('left', 'right')
                    if _is_caller_left_side(action.caller, action.callee) else
                    ('right', 'left'))
                note_callee, note_caller = action.notes
                if note_callee:
                    _generate_note(output, callee_side, obj=None,
                                   text=note_callee)
                if note_caller:
                    _generate_note(output, caller_side, obj=None,
                                   text=note_caller)

            if not isinstance(n_action, sd_action.ImplicitReturn):
                output.append('activate %s' % action.callee)
            call_stack.append(current_call)
            current_call = action

        elif isinstance(action, sd_action.ImplicitReturn):
            if not isinstance(p_action, sd_action.Call):
                output.append('deactivate %s' % current_call.callee)

            if 'd' in current_call.flags:
                output.append('destroy %s' % current_call.callee)

            current_call = call_stack.pop()

        elif isinstance(action, sd_action.Return):
            s = '%s <-- %s' % (current_call.caller, current_call.callee)
            params = str(action.params)
            if params:
                s += ': %s' % action.params
            output.append(s)
            output.append('deactivate %s' % current_call.callee)
            current_call = call_stack.pop()

        elif isinstance(action, sd_action.FragBegin):
            if action.op_name == 'alt':
                is_alt_waiting_for_first_choice = True
            else:
                if action.op_name == 'choice':
                    if is_alt_waiting_for_first_choice:
                        is_alt_waiting_for_first_choice = False
                        s = 'alt'
                    else:
                        s = 'else'
                else:
                    s = '%s' % (action.op_name)

                if action.condition:
                    s += ' %s' % action.condition
                output.append(s)

        elif isinstance(action, sd_action.FragEnd):
            if action.op_name == 'choice':
                pass
            else:
                output.append('end')

        elif isinstance(action, sd_action.Note):
            _generate_note(output, None, action.obj, action.text)

        elif isinstance(action, sd_action.Delay):
            if action.text:
                output.append('... {} ...'.format(action.text))
            else:
                output.append('...')

    output.append('@enduml\n')
    return '\n'.join(output)


def generate(diagram_name, output_dir, sd_context, _=None):
    script = _generate_script(sd_context)
    output_path = os.path.join(output_dir, diagram_name + '.puml')
    with open(output_path, 'wt') as f:
        f.write(script)
    return [output_path]



"""
Actions to be captured as sequence.
"""


class _Action(object):
    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class Call(_Action):
    def __init__(self, caller, callee, method_name, params,
                 flags='', notes=None):
        self.caller = caller
        self.callee = callee
        self.method_name = method_name
        self.params = params
        self.flags = flags
        self.notes = notes if notes else [None, None]

    def __repr__(self):
        return 'call from %s to %s::%s(%s) [%s%s]' % (
            self.caller,
            self.callee,
            self.method_name,
            self.params,
            self.flags,
            self.notes)

    def __eq__(self, other):
        return (self.caller is other.caller and
                self.caller is other.caller and
                self.method_name == other.method_name and
                self.params == other.params and
                self.flags == other.flags,
                self.notes == other.notes)


class Return(_Action):
    def __init__(self, params):
        self.params = params

    def __repr__(self):
        return 'return (%s)' % (self.params)

    def __eq__(self, other):
        return self.params == other.params


class ImplicitReturn(_Action):
    def __init__(self):
        pass

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def __repr__(self):
        return 'implicit return'


class FragBegin(_Action):
    def __init__(self, op_name, condition=None):
        self.op_name = op_name
        self.condition = condition

    def __repr__(self):
        s = '%s begin' % self.op_name
        if self.condition:
            s += ' [%s]' % self.condition
        return s

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.__dict__ == other.__dict__)


class FragEnd(_Action):
    def __init__(self, op_name):
        self.op_name = op_name

    def __repr__(self):
        return '%s end' % self.op_name

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.__dict__ == other.__dict__)


class Note(_Action):
    def __init__(self, text, obj=None):
        self.text = text
        self.obj = obj

    def __repr__(self):
        return 'note over %s : %s' % (self.obj, self.text)

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.__dict__ == other.__dict__)


class Delay(_Action):
    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return 'delay' + ('({})'.foramt(self.text) if self.text else '')

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.text == other.text)



"""
Generate PlantUML script and image file
"""
import os
import string
import base64
import zlib
import requests

from . import gen_plantuml

DEFAULT_SERVER_URL = 'http://www.plantuml.com/plantuml'

_BASE64_TO_PLANTUML = {ord(b): b2.encode() for b, b2 in zip(
    string.ascii_uppercase + string.ascii_lowercase + string.digits + '+/=',
    string.digits + string.ascii_uppercase + string.ascii_lowercase + '-_=')}


def _encode_text_diagram(text_diagram):
    """
    Encode text diagram with zlib/plantuml specific b64 encoding.

    The text diagram is:
    - Encoded in UTF-8
    - Compressed using Deflate algorithm
    - Re-encoded in ASCII using a transformation close to base64

    See https://plantuml.com/text-encoding:
    """
    utf_encoded = text_diagram.encode('utf-8')
    compressed = zlib.compress(utf_encoded)
    # Remove zlib header CMF/CM[2 bytes] and CRC[4 bytes] (RFC1950). Note that
    # Python zlib retains those but Java.util.zip.Inflater from PlantUML server
    # does not expect them.
    compressed = compressed[2:-4]
    b64_encoded = base64.b64encode(compressed)
    return b''.join(_BASE64_TO_PLANTUML[b] for b in b64_encoded)


def _get_image_type(output_image_file_name):
    _, ext = os.path.splitext(output_image_file_name)
    assert ext.startswith('.')
    return ext[1:]


def _get_server_url(server_url):
    if not server_url:
        return DEFAULT_SERVER_URL
    return server_url[:-1] if server_url.endswith('/') else server_url


def generate_image(plantuml_file_path, image_file_path, server_url=None):
    """
    Generate image file from plantuml text file using server.

    The type of image is determined by the extension name of image_file_path.
    Default public PlantUML server is used if server_url is None.
    """
    with open(plantuml_file_path, 'rt') as input_file:
        text_diagram = input_file.read()

    encoded_diagram = _encode_text_diagram(text_diagram)
    diagram_url = encoded_diagram.decode('utf-8')

    image_type = _get_image_type(image_file_path)
    server_url = _get_server_url(server_url)
    url = server_url + "/" + image_type + "/" + diagram_url
    response = requests.get(url)

    if response.status_code == 200:
        with open(image_file_path, 'wb') as output_file:
            output_file.write(response.content)
    else:
        response.raise_for_status()


def generate(diagram_name, output_dir, sd_context, options, image_type):
    """
    Generate both plantuml file and image file.
    """
    generated_files = gen_plantuml.generate(diagram_name,
                                            output_dir, sd_context, options)
    plantuml_file_path = generated_files[0]
    image_path = os.path.join(output_dir, diagram_name + '.' + image_type)
    generate_image(plantuml_file_path, image_path, options.get('server_url'))

    generated_files.append(image_path)
    return generated_files



from napkin import util


class TestNeighbour(object):
    def test_empty_item(self):
        it = util.neighbour([])
        assert [] == list(it)

    def test_one_item(self):
        it = util.neighbour([1])
        assert list(it) == [(None, 1, None)]

    def test_two_items(self):
        it = util.neighbour([1, 2])
        assert list(it) == [(None, 1, 2), (1, 2, None)]

    def test_three_items(self):
        it = util.neighbour([1, 2, 3])
        assert list(it) == [(None, 1, 2), (1, 2, 3), (2, 3, None)]



import pytest
from napkin import gen_plantuml
from napkin import sd


@pytest.fixture
def check(tmpdir):
    def fn(sd_func, exp_lines):
        exp_lines = '@startuml' + exp_lines + '@enduml\n'
        sd_context = sd.parse(sd_func)
        puml_file = gen_plantuml.generate(str(tmpdir), 'test', sd_context)[0]
        with open(puml_file, 'rt') as f:
            lines = f.read()
        assert lines == exp_lines
    return fn


class TestSimpleCalls:
    def test_call(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func()

        check(f, """
participant foo
participant bar

foo -> bar : func()
""")

    def test_call_with_params(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func('abc')
                bar.func2(key='value')

        check(f, """
participant foo
participant bar

foo -> bar : func(abc)
foo -> bar : func2(key=value)
""")

    def test_call_with_return(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func().ret()

        check(f, """
participant foo
participant bar

foo -> bar : func()
activate bar
foo <-- bar
deactivate bar
""")

    def test_call_twice(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            baz = c.object('baz')
            with foo:
                bar.func()
                baz.func()

        check(f, """
participant foo
participant bar
participant baz

foo -> bar : func()
foo -> baz : func()
""")

    def test_call_two_level(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            baz = c.object('baz')
            with foo:
                with bar.func():
                    baz.func()

        check(f, """
participant foo
participant bar
participant baz

foo -> bar : func()
activate bar
bar -> baz : func()
deactivate bar
""")

    def test_opt(self, check):
        def f(c):
            foo = c.object('foo')
            baz = c.object('baz')
            with foo:
                with c.opt():
                    baz.func()

        check(f, """
participant foo
participant baz

opt
foo -> baz : func()
end
""")

    def test_opt_with_condition(self, check):
        def f(c):
            foo = c.object('foo')
            baz = c.object('baz')
            with foo:
                with c.opt('is_ok'):
                    baz.func()

        check(f, """
participant foo
participant baz

opt is_ok
foo -> baz : func()
end
""")

    def test_alt(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            baz = c.object('baz')
            with foo:
                with c.alt():
                    with c.choice('a'):
                        baz.func()
                    with c.choice('b'):
                        bar.func()

        check(f, """
participant foo
participant bar
participant baz

alt a
foo -> baz : func()
else b
foo -> bar : func()
end
""")


class TestObjectWithClassAndStereotype:
    def test_call(self, check):
        def f(c):
            a = c.object('a', cls='Foo')
            b = c.object('b', cls='Foo', stereotype='foo')
            with a:
                b.func()

        check(f, """
participant "a:Foo" as a
participant "b:Foo" as b <<foo>>

a -> b : func()
""")


class TestCreate:
    def test_simple_call(self, check):
        def f(c):
            foo = c.object('foo')
            with foo:
                bar = c.object('bar')
                c.create(bar)

        check(f, """
participant foo
participant bar

create bar
foo -> bar : <<create>>
""")

    def test_call_in_constructor(self, check):
        def f(c):
            foo = c.object('foo')
            baz = c.object('baz')
            with foo:
                bar = c.object('bar')
                with c.create(bar):
                    baz.func()

        check(f, """
participant foo
participant baz
participant bar

create bar
foo -> bar : <<create>>
activate bar
bar -> baz : func()
deactivate bar
""")


class TestDestroy:
    def test_simple_destroy(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func()
                c.destroy(bar)

        check(f, """
participant foo
participant bar

foo -> bar : func()
foo -> bar : <<destroy>>
destroy bar
""")

    def test_call_in_destructor(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func()
                with c.destroy(bar):
                    foo.end()

        check(f, """
participant foo
participant bar

foo -> bar : func()
foo -> bar : <<destroy>>
activate bar
bar -> foo : end()
deactivate bar
destroy bar
""")


class TestNote:
    def test_over_object(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                c.note('blah')
                bar.func()

        check(f, """
participant foo
participant bar

note over foo : blah
foo -> bar : func()
""")

    def test_multiple_line_text(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                c.note('blah\nblah')
                bar.func()

        check(f, """
participant foo
participant bar

note over foo
blah
blah
end note
foo -> bar : func()
""")

    def test_callee_caller(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func().note(callee='callee side', caller='caller side')

        check(f, """
participant foo
participant bar

foo -> bar : func()
note right : callee side
note left : caller side
""")


class TestDelay:
    def test_without_text(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func()
            c.delay()

        check(f, """
participant foo
participant bar

foo -> bar : func()
...
""")

    def test_with_text(self, check):
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func()
            c.delay('hello')

        check(f, """
participant foo
participant bar

foo -> bar : func()
... hello ...
""")

    def test_group(self, check):
        def f(c):
            foo = c.object('foo')
            baz = c.object('baz')
            with foo:
                with c.group():
                    baz.func()

        check(f, """
participant foo
participant baz

group
foo -> baz : func()
end
""")

    def test_group_with_label(self, check):
        def f(c):
            foo = c.object('foo')
            baz = c.object('baz')
            with foo:
                with c.group('label'):
                    baz.func()

        check(f, """
participant foo
participant baz

group label
foo -> baz : func()
end
""")



import os
import napkin


class TestDecorator(object):
    def test_seq_diagram(self, tmpdir):
        fname = os.path.join(str(tmpdir), 'sd')

        @napkin.seq_diagram(fname)
        def f(c):
            foo = c.object('foo')
            bar = c.object('bar')
            with foo:
                bar.func()
        napkin.generate()

        # Check diagram generate with fname
        assert os.stat(fname + '.puml')



import setuptools
import re

with open('README.md', 'rt') as f:
    long_description = f.read()

with open('napkin/__init__.py', "rt") as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setuptools.setup(
    name='napkin',
    version=version,
    author='Hans Jang',
    author_email='hsjang8848@gmail.com',
    description='Python DSL for writing PlantUML sequence diagram',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/pinetr2e/napkin',
    packages=setuptools.find_packages(),
    install_requires=['requests'],
    setup_requires=['pytest-runner'],
    extras_require={
        'PNG': ['plantuml'],
    },
    tests_require=["pytest"],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Software Development',
    ],
    entry_points={
        'console_scripts': [
            'napkin = napkin.cli:main',
            'napkin_plantuml = napkin.plantuml_cli:main',
        ],
    },
)



import pytest

from napkin import sd
from napkin import sd_action


class TestParams:
    def test_empty(self):
        assert "" == str(sd.Params(tuple(), dict()))

    def test_args(self):
        assert "abc, def" == str(sd.Params(('abc', 'def'), dict()))

    def test_args_kargs(self):
        assert "abc, foo=1" == str(sd.Params(('abc',),
                                             dict(foo=1)))


class TestBase(object):
    def check(self, context, exp_actions):
        actions = context._sequence

        # This is for better debugging
        assert str(actions) == str(exp_actions)
        assert actions == exp_actions


class TestTopLevel(TestBase):
    def test_call(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
        ])

    def test_call_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()
            bar.func2()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.Call(foo, bar, 'func2', sd.Params()),
            sd_action.ImplicitReturn(),
        ])

    def test_call_with_return(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func().ret('val')

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.Return(sd.Params(('val',))),
        ])

    def test_call_with_return_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func().ret('val')
            bar.func2().ret('val2')

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.Return(sd.Params(('val',))),
            sd_action.Call(foo, bar, 'func2', sd.Params()),
            sd_action.Return(sd.Params(('val2',))),
        ])

    def test_fail_when_separate_return_called(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()
            with pytest.raises(sd.CallError):
                c.ret()

    def test_fail_when_top_level_caller_set_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()
            with pytest.raises(sd.TopLevelCallerError):
                with foo:
                    bar.func()

    def test_noop_when_do_nothing_in_top_level_caller(self):
        c = sd.Context()

        foo = c.object('foo')
        with foo:
            pass

        self.check(c, [
        ])


class TestSecondLevel(TestBase):
    def test_call(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            with bar.func():
                baz.func2()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.Call(bar, baz, 'func2', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.ImplicitReturn(),
        ])

    def test_call_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            with bar.func():
                baz.func2()
                baz.func3()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.Call(bar, baz, 'func2', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.Call(bar, baz, 'func3', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.ImplicitReturn(),
        ])

    def test_call_with_return(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            with bar.func():
                baz.func2().ret()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.Call(bar, baz, 'func2', sd.Params()),
            sd_action.Return(sd.Params()),
            sd_action.ImplicitReturn(),
        ])

    def test_return_from_outside_func(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            with bar.func():
                baz.func2()
                c.ret()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.Call(bar, baz, 'func2', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.Return(sd.Params()),
        ])

    def test_fail_when_call_after_returning_from_outside_func(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            with bar.func():
                baz.func2()
                c.ret()
                with pytest.raises(sd.CallError):
                    baz.func3()

    def test_fail_when_return_again_from_outside_func(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            with bar.func():
                baz.func2()
                c.ret()
                with pytest.raises(sd.ReturnError):
                    c.ret()

    def test_return_from_outside_func_without_calling_any(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            with bar.func():
                c.ret()

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.Return(sd.Params()),
        ])

    def test_do_nothing_in_outside_func(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            with bar.func():
                pass

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
        ])


class TestCreate(TestBase):
    def test_simple(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            c.create(bar)

        self.check(c, [
            sd_action.Call(foo, bar, '<<create>>', sd.Params(), flags='c'),
            sd_action.ImplicitReturn(),
        ])

    def test_non_default_method(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            c.create(bar.new())

        self.check(c, [
            sd_action.Call(foo, bar, 'new', sd.Params(), flags='c'),
            sd_action.ImplicitReturn(),
        ])

    def test_constructor_params(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            c.create(bar.new('a', name='bar'))

        self.check(c, [
            sd_action.Call(foo, bar, 'new',
                           params=sd.Params(('a',), dict(name='bar')),
                           flags='c'),
            sd_action.ImplicitReturn(),
        ])

    def test_call_others_in_constructor(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            with c.create(bar):
                baz.func()

        self.check(c, [
            sd_action.Call(foo, bar, '<<create>>', sd.Params(), flags='c'),
            sd_action.Call(bar, baz, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.ImplicitReturn(),
        ])

    def test_fail_if_called_at_top_level(self):
        c = sd.Context()

        with pytest.raises(sd.CreateError):
            bar = c.object('bar')
            c.create(bar)

    def test_fail_if_create_object_already_being_used(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()
            with pytest.raises(sd.CreateError):
                c.create(bar)

    def test_fail_if_create_object_twice(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            c.create(bar)
            with pytest.raises(sd.CreateError):
                c.create(bar)


class TestDestroy(TestBase):
    def test_simple(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            bar.func()
            c.destroy(bar)

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
            sd_action.Call(foo, bar, '<<destroy>>', sd.Params(), flags='d'),
            sd_action.ImplicitReturn(),
        ])

    def test_fail_when_call_destroyed_object(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            c.destroy(bar)
            with pytest.raises(sd.CallError):
                bar.func()

    def test_call_other_methods_of_the_same_object_from_destructr(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            with c.destroy(bar):
                bar.func()

    def test_fail_when_destroy_twice_the_same_object(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            c.destroy(bar)
            with pytest.raises(sd.CallError):
                c.destroy(bar)

    def test_fail_if_called_at_top_level(self):
        c = sd.Context()

        foo = c.object('foo')
        with pytest.raises(sd.DestroyError):
            c.destroy(foo)


class TestNote(TestBase):
    def test_over_object_implicit(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            c.note('blah')
            bar.func()

        self.check(c, [
            sd_action.Note('blah', obj=foo),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
        ])

    def test_over_object_explicit(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        with foo:
            foo.note('blah')
            bar.note('blah2')
            bar.func()

        self.check(c, [
            sd_action.Note('blah', obj=foo),
            sd_action.Note('blah2', obj=bar),
            sd_action.Call(foo, bar, 'func', sd.Params()),
            sd_action.ImplicitReturn(),
        ])

    def test_call_specific(self):
        c = sd.Context()

        foo = c.object('foo')
        bar = c.object('bar')
        baz = c.object('baz')
        with foo:
            bar.func().note('callee side note')
            baz.func().note(caller='caller side note',
                            callee='callee side note')
            baz.func2().note('note').ret('val')

        self.check(c, [
            sd_action.Call(foo, bar, 'func', sd.Params(),
                           notes=['callee side note', None]),
            sd_action.ImplicitReturn(),
            sd_action.Call(foo, baz, 'func', sd.Params(),
                           notes=['callee side note', 'caller side note']),
            sd_action.ImplicitReturn(),
            sd_action.Call(foo, baz, 'func2', sd.Params(),
                           notes=['note', None]),
            sd_action.Return(sd.Params(('val',))),
        ])



import itertools


def neighbour(seq):
    """
    Generate sequence of adjacent items.
    For example, [1, 2] -> (None, 1, 2), (1, 2, None)
    """
    it = iter(seq)
    it_next = itertools.islice(itertools.chain(iter(seq), [None]), 1, None)

    prev = None
    for curr, next in zip(it, it_next):
        yield(prev, curr, next)
        prev = curr



"""
Generate PlantUML script and PNG file
"""
from . import gen_plantuml_img


def generate(diagram_name, output_dir, sd_context, options=None):
    return gen_plantuml_img.generate(diagram_name, output_dir, sd_context,
                                     options, 'svg')
